package com.sun.org.apache.bcel.internal.classfile;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.DataInput;

public final class Code extends Attribute
{
    private int max_stack;
    private int max_locals;
    private byte[] code;
    private CodeException[] exception_table;
    private Attribute[] attributes;
    
    public Code(Code c) {
        this(c.getNameIndex(), c.getLength(), c.getMaxStack(), c.getMaxLocals(), c.getCode(), c.getExceptionTable(), c.getAttributes(), c.getConstantPool());
    }
    
    Code(int name_index, int length, DataInput file, ConstantPool constant_pool) throws IOException {
        this(name_index, length, file.readUnsignedShort(), file.readUnsignedShort(), (byte[])null, (CodeException[])null, (Attribute[])null, constant_pool);
        int code_length = file.readInt();
        file.readFully(this.code = new byte[code_length]);
        int exception_table_length = file.readUnsignedShort();
        this.exception_table = new CodeException[exception_table_length];
        for (int i = 0; i < exception_table_length; ++i) {
            this.exception_table[i] = new CodeException(file);
        }
        int attributes_count = file.readUnsignedShort();
        this.attributes = new Attribute[attributes_count];
        for (int j = 0; j < attributes_count; ++j) {
            this.attributes[j] = Attribute.readAttribute(file, constant_pool);
        }
        super.setLength(length);
    }
    
    public Code(int name_index, int length, int max_stack, int max_locals, byte[] code, CodeException[] exception_table, Attribute[] attributes, ConstantPool constant_pool) {
        super((byte)2, name_index, length, constant_pool);
        this.max_stack = max_stack;
        this.max_locals = max_locals;
        this.code = ((code != null) ? code : new byte[0]);
        this.exception_table = ((exception_table != null) ? exception_table : new CodeException[0]);
        this.attributes = ((attributes != null) ? attributes : new Attribute[0]);
        super.setLength(this.calculateLength());
    }
    
    @Override
    public void accept(Visitor v) {
        v.visitCode(this);
    }
    
    @Override
    public final void dump(DataOutputStream file) throws IOException {
        super.dump(file);
        file.writeShort(this.max_stack);
        file.writeShort(this.max_locals);
        file.writeInt(this.code.length);
        file.write(this.code, 0, this.code.length);
        file.writeShort(this.exception_table.length);
        for (CodeException exception : this.exception_table) {
            exception.dump(file);
        }
        file.writeShort(this.attributes.length);
        for (Attribute attribute : this.attributes) {
            attribute.dump(file);
        }
    }
    
    public final Attribute[] getAttributes() {
        return this.attributes;
    }
    
    public LineNumberTable getLineNumberTable() {
        Attribute[] attributes = this.attributes;
        int length = attributes.length;
        int i = 0;
        while (i < length) {
            Attribute attribute = attributes[i];
            if ((attribute instanceof LineNumberTable)) {
                return (LineNumberTable)attribute;
            }
            ++i;
        }
        return null;
    }
    
    public LocalVariableTable getLocalVariableTable() {
        Attribute[] attributes = this.attributes;
        int length = attributes.length;
        int i = 0;
        while (i < length) {
            Attribute attribute = attributes[i];
            if ((attribute instanceof LocalVariableTable)) {
                return (LocalVariableTable)attribute;
            }
            ++i;
        }
        return null;
    }
    
    public final byte[] getCode() {
        return this.code;
    }
    
    public final CodeException[] getExceptionTable() {
        return this.exception_table;
    }
    
    public final int getMaxLocals() {
        return this.max_locals;
    }
    
    public final int getMaxStack() {
        return this.max_stack;
    }
    
    private int getInternalLength() {
        return 8 + this.code.length + 2 + 8 * ((this.exception_table == null) ? 0 : this.exception_table.length) + 2;
    }
    
    private int calculateLength() {
        int len = 0;
        if (this.attributes != null) {
            for (Attribute attribute : this.attributes) {
                len += attribute.getLength() + 6;
            }
        }
        return len + this.getInternalLength();
    }
    
    public final void setAttributes(Attribute[] attributes) {
        this.attributes = ((attributes != null) ? attributes : new Attribute[0]);
        super.setLength(this.calculateLength());
    }
    
    public final void setCode(byte[] code) {
        this.code = ((code != null) ? code : new byte[0]);
        super.setLength(this.calculateLength());
    }
    
    public final void setExceptionTable(CodeException[] exception_table) {
        this.exception_table = ((exception_table != null) ? exception_table : new CodeException[0]);
        super.setLength(this.calculateLength());
    }
    
    public final void setMaxLocals(int max_locals) {
        this.max_locals = max_locals;
    }
    
    public final void setMaxStack(int max_stack) {
        this.max_stack = max_stack;
    }
    
    public final String toString(boolean verbose) {
        StringBuilder buf = new StringBuilder(100);
        buf.append("Code(max_stack = ").append(this.max_stack).append(", max_locals = ").append(this.max_locals).append(", code_length = ").append(this.code.length).append(")\n").append(Utility.codeToString(this.code, super.getConstantPool(), 0, -1, verbose));
        if (this.exception_table.length > 0) {
            buf.append("\nException handler(s) = \n").append("From\tTo\tHandler\tType\n");
            for (CodeException exception : this.exception_table) {
                buf.append(exception.toString(super.getConstantPool(), verbose)).append("\n");
            }
        }
        if (this.attributes.length > 0) {
            buf.append("\nAttribute(s) = ");
            for (Attribute attribute : this.attributes) {
                buf.append("\n").append(attribute);
            }
        }
        return buf.toString();
    }
    
    @Override
    public final String toString() {
        return this.toString(true);
    }
    
    @Override
    public Attribute copy(ConstantPool _constant_pool) {
        Code c = (Code)this.clone();
        if (this.code != null) {
            c.code = new byte[this.code.length];
            System.arraycopy(this.code, 0, c.code, 0, this.code.length);
        }
        c.setConstantPool(_constant_pool);
        c.exception_table = new CodeException[this.exception_table.length];
        for (int i = 0; i < this.exception_table.length; ++i) {
            c.exception_table[i] = this.exception_table[i].copy();
        }
        c.attributes = new Attribute[this.attributes.length];
        for (int i = 0; i < this.attributes.length; ++i) {
            c.attributes[i] = this.attributes[i].copy(_constant_pool);
        }
        return c;
    }
}
